------------------------------------------------------
-- AutoCraft.lua
-- Version 1700.2
-- Originally based on BetterTradeSkill by Matt Perry (http://somewhere.fscked.org)
------------------------------------------------------

FAC_Config = {};
FAC_Config.AutoIterate = false;

local P = {};
AutoCraft = P;

P.OrderedQueue = {};
P.QueuedReagents = {};
P.ManuallyPaused = false;
P.SkillCache = {};
P.ReagentCache = {};

local MAX_QUEUE_BUTTONS = 10;
local CACHE_TIME = 1;	-- We cache the returns from our hooked GetTradeSkillInfo because it can be called many times a second without actually having any relevant conditions be different. This keeps us from repeatedly recalculating it when not needed.
local QUEUE_ADVANCE_DELAY = 1; -- In 1.8 we seem to have problems if we try to advance the queue too soon after finishing an item (if the item we just finished is needed to craft the next item in the queue, WoW thinks we don't have it yet and thus can't craft the next item). May have something to do with lag. So, we wait this long before trying to advance the queue.


function P.GetTradeSkillInfo(index)
	local skillName, difficulty, numAvailable, isExpanded = P.Orig_GetTradeSkillInfo(index);
	local itemName = P.GetTradeItemName(index);
	
	if (P.SkillCache[skillName] and (GetTime() - P.SkillCache[skillName].time < CACHE_TIME)) then
		local info = P.SkillCache[skillName].info;
		if (info.skillName == skillName and info.difficulty == difficulty and info.numAvailable == numAvailable and info.isExpanded == isExpanded) then
			return info.skillName, info.difficulty, info.numAvailable, info.isExpanded, itemName;
		end
	end
	
	-- Adjustment #1: recalculate numAvailable based on our altered GetTradeSkillReagentInfo (normally its value is calculated server-side)
	local reagentTally = {};
	for reagentIndex = 1, GetTradeSkillNumReagents(index) do
		local reagentName, _, reagentCount, playerReagentCount = P.GetTradeSkillReagentInfo(index, reagentIndex);
		local completeItemsWorth = math.max(math.floor(playerReagentCount / reagentCount), 0); 
		table.insert(reagentTally, completeItemsWorth); -- how many could we make if this were the only reagent in question?
	end
	table.sort(reagentTally); -- first item is the minimum of the set once it's sorted
	numAvailable = (reagentTally[1] or 0); -- this is how many we can really make if we include already-queued reagents
	
	-- Adjustment #2 (TODO): how many of this item could we make if we were to make/buy the reagents right now?
		
	if (skillName ~= nil) then
		P.SkillCache[skillName] = { time=GetTime(), info={ skillName=skillName, difficulty=difficulty, numAvailable=numAvailable, isExpanded=isExpanded } };
	end
	
	return skillName, difficulty, numAvailable, isExpanded, itemName;
end

function P.GetTradeSkillReagentInfo(skillIndex, reagentIndex)
	local reagentName, reagentTexture, reagentCount, playerReagentCount = P.Orig_GetTradeSkillReagentInfo(skillIndex, reagentIndex);
	
	if (P.ReagentCache[reagentName] and (GetTime() - P.ReagentCache[reagentName].time < CACHE_TIME)) then
		local info = P.ReagentCache[reagentName].info;
		if (info.reagentTexture == reagentTexture and info.reagentCount == reagentCount and info.playerReagentCount == playerReagentCount) then
			return reagentName, info.reagentTexture, info.reagentCount, info.playerReagentCount;
		end
	end

	-- Adjustment #1: if you've queued some of this item, add them to effective "inventory" for determining what you can make from it.
	playerReagentCount = playerReagentCount + P.NumOfItemInQueue(reagentName);

	-- Adjustment #2: if your queue contains items that use this reagent, subtract the amount that will be used from your "inventory".
	for _, queueEntry in P.OrderedQueue do
		local perItemCount = queueEntry.reagentsPerItem[reagentName];
		if (perItemCount ~= nil) then
			playerReagentCount = playerReagentCount - ((queueEntry.numQueued - queueEntry.numProduced) * perItemCount);
		end
	end
	playerReagentCount = math.max(playerReagentCount, 0);
	
	if (reagentName ~= nil) then
		P.ReagentCache[reagentName] = { time=GetTime(), info={ reagentTexture=reagentTexture, reagentCount=reagentCount, playerReagentCount=playerReagentCount } };
	end

	return reagentName, reagentTexture, reagentCount, playerReagentCount;
end

function P.GetTradeSkillTools(skillIndex)
	local toolInfo = {P.Orig_GetTradeSkillTools(skillIndex)};
	local newToolInfo = {};
	for index = 1, table.getn(toolInfo) do
		local item = toolInfo[index];
		if (item and type(item) == "string") then
			table.insert(newToolInfo, item);
			if (toolInfo[index+1] == 1 or P.NumOfItemInQueue(item) > 0) then
				table.insert(newToolInfo, 1);
			end
		end
	end
	return unpack(newToolInfo);
end

function P.NumOfItemInQueue(itemName)
	for _, queueEntry in P.OrderedQueue do
		if (queueEntry.item == itemName) then
			return (queueEntry.numQueued or 0);
		end
	end
	return 0;
end

function P.TradeSkillFrame_SetSelection(index)
	P.Orig_TradeSkillFrame_SetSelection(index);

	local skillName, difficulty, numAvailable, isExpanded = P.GetTradeSkillInfo(index);	
	if (numAvailable and numAvailable > 0) then
		AutoCraftQueue:Enable();
		AutoCraftQueueAll:Enable();
	else
		AutoCraftQueue:Disable();
		AutoCraftQueueAll:Disable();
	end
end

function P.TradeSkillFrame_LoadUI()
	P.Orig_TradeSkillFrame_LoadUI();

	if (not P.Loaded) then
		-- Attach our UI elements to the TradeSkillFrame now that it's loaded.
		AutoCraftBackground:SetParent(TradeSkillFrame);
		AutoCraftBackground:SetPoint("TOPLEFT", 0, 0);	
		AutoCraftBackground:SetHeight(TradeSkillFrame:GetHeight() + 48); 
			-- reset this in case someone (say, EnhancedTradeSkill) resized TradeSkillFrame on us
		AutoCraftBackground:Show();	
		AutoCraftQueue:SetParent(TradeSkillFrame);	
		AutoCraftQueue:SetPoint("TOPLEFT", TradeSkillCreateButton, "TOPLEFT", 0, 0);	
		AutoCraftQueue:Show();	
		AutoCraftQueueAll:SetParent(TradeSkillFrame);	
		AutoCraftQueueAll:SetPoint("TOPLEFT", TradeSkillCreateAllButton, "TOPLEFT", 0, 0);	
		AutoCraftQueueAll:Show();	
		AutoCraftBuy:SetParent(TradeSkillFrame);	
		AutoCraftBuy:SetPoint("TOPLEFT", TradeSkillCancelButton, "TOPLEFT", 0, 0);	
		AutoCraftBuy:Show();	
		AutoCraftQueueEverything:SetParent(TradeSkillFrame);	
		AutoCraftQueueEverything:Show();	
		AutoCraftRunQueue:SetParent(TradeSkillFrame);	
		AutoCraftRunQueue:Show();	
		AutoCraftClear:SetParent(TradeSkillFrame);	
		AutoCraftClear:Show();	
		AutoCraftRunAutomatically:SetParent(TradeSkillFrame);	
		AutoCraftRunAutomatically:Show();	
		AutoCraftRunAutomaticallyText:SetText(FAC_RUN_AUTOMATICALLY_CHECKBUTTON);
		AutoCraftRunAutomatically:SetChecked(FAC_Config.AutoIterate);

		-- Hide the old Create buttons so I can put my buttons in  their place.  
		-- Not sure if this will always work, but it does at the moment.
		TradeSkillCreateButton:Hide();
		TradeSkillCreateAllButton:Hide();
		TradeSkillCancelButton:Hide();
				
		P.Orig_GetTradeSkillTools = GetTradeSkillTools;
		GetTradeSkillTools = P.GetTradeSkillTools;
		P.Orig_GetTradeSkillReagentInfo = GetTradeSkillReagentInfo;
		GetTradeSkillReagentInfo = P.GetTradeSkillReagentInfo;
		P.Orig_GetTradeSkillInfo = GetTradeSkillInfo;
		GetTradeSkillInfo = P.GetTradeSkillInfo;
		P.Orig_TradeSkillFrame_SetSelection = TradeSkillFrame_SetSelection;
		TradeSkillFrame_SetSelection = P.TradeSkillFrame_SetSelection;
		
		P.Loaded = true;
		GFWUtils.Print("Fizzwidget AutoCraft loaded!");
	end
end

function P.OnUpdate()
	if (P.AdvanceQueueTime ~= nil and GetTime() >= P.AdvanceQueueTime) then
		P.Iterate();
		P.AdvanceQueueTime = nil;
	end
end

function P.OnLoad()
	-- we do all our loading once the TradeSkillFrame loads (which is on-demand now).
	P.Orig_TradeSkillFrame_LoadUI = TradeSkillFrame_LoadUI;
	TradeSkillFrame_LoadUI = P.TradeSkillFrame_LoadUI;

	this:RegisterEvent("CHAT_MSG_SPELL_TRADESKILLS");
	this:RegisterEvent("MERCHANT_SHOW");
	this:RegisterEvent("MERCHANT_CLOSED");
	this:RegisterEvent("TRADE_SKILL_SHOW");
	this:RegisterEvent("TRADE_SKILL_CLOSE");
	this:RegisterEvent("TRADE_SKILL_UPDATE");
	this:RegisterEvent("SPELLCAST_INTERRUPTED");
	this:RegisterEvent("SPELLCAST_FAILED");
	this:RegisterEvent("SPELLCAST_START");
	this:RegisterEvent("SPELLCAST_STOP");

end

function P.OnEvent(event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
	if (not P.Loaded) then 
		-- all the events we care about only have relevance once we're fully in action and attached to the TradeSkill UI
		return; 
	end
	if (event == "CHAT_MSG_SPELL_TRADESKILLS") then
		if (FAC_CreateItemPattern == nil) then
			FAC_CreateItemPattern = GFWUtils.FormatToPattern(TRADESKILL_LOG_FIRSTPERSON);
		end
		for item in string.gfind(arg1, FAC_CreateItemPattern) do
			P.HandleCreated(item);
			return;
		end
	elseif (event == "TRADE_SKILL_CLOSE" or event == "TRADE_SKILL_UPDATE" or event == "MERCHANT_SHOW" or event == "MERCHANT_CLOSED") then
		P.UpdateDisplay();
	elseif (event == "TRADE_SKILL_SHOW") then
		P.UpdateDisplay();
		AutoCraftRunAutomaticallyText:SetText(FAC_RUN_AUTOMATICALLY_CHECKBUTTON);
		AutoCraftRunAutomatically:SetChecked(FAC_Config.AutoIterate);
		-- if we've just switched to the tradeskill window needed for the head of the queue, start iterating again
		if (not P.ManuallyPaused and P.OrderedQueue[1] ~= nil and P.OrderedQueue[1].skill == GetTradeSkillLine()) then
			P.Iterate();
		end
	elseif (event == "SPELLCAST_FAILED" or event == "SPELLCAST_INTERRUPTED") then
		if (P.Unqueueing or P.CastingOtherSpell) then
			P.Unqueueing = false; 
			P.CastingOtherSpell = false;
			P.Iterate();
			return;
		end
		P.CastingOtherSpell = false;
		P.CraftingQueuedItem = false;
		P.UpdateDisplay();
	elseif (event == "SPELLCAST_STOP") then
		if (P.CastingOtherSpell) then
			P.CastingOtherSpell = false;
			P.Iterate();
			P.UpdateDisplay();
		end
	elseif (event == "SPELLCAST_START") then
		if (P.OrderedQueue[1] and CastingBarText:GetText() == P.OrderedQueue[1].recipe) then
			P.CraftingQueuedItem = true;
			P.CastingOtherSpell = false;
		else
			P.CraftingQueuedItem = false;
			P.CastingOtherSpell = true;
		end
		P.UpdateDisplay();
	end
end

function P.Queue_OnClick()
	local id = tonumber(this:GetID());
	if (P.OrderedQueue[id]) then
		table.remove(P.OrderedQueue, id);
		if (id == 1 and P.CraftingQueuedItem) then
			P.Unqueueing = true;
			SpellStopCasting();
		end
		TradeSkillFrame_SetSelection(GetTradeSkillSelectionIndex());
		TradeSkillFrame_Update();	
		P.UpdateDisplay();
		P.Queue_UpdateTooltip(id);
	end
end

function P.Queue_OnEvent()
	P.Queue_UpdateTooltip(tonumber(this:GetID()));
end

function P.Queue_UpdateTooltip(id)
	local button = getglobal("QueuedSkillIcon"..id);
	if (button:IsVisible()) then
		GameTooltip:SetOwner(button, "ANCHOR_TOPLEFT");
		if ( id == 1 and P.CraftingQueuedItem) then
			GameTooltip:SetText(NORMAL_FONT_COLOR_CODE..FAC_CREATING..FONT_COLOR_CODE_CLOSE, 1.0, 1.0, 1.0);
		else
			GameTooltip:SetText(NORMAL_FONT_COLOR_CODE..FAC_QUEUED..FONT_COLOR_CODE_CLOSE, 1.0, 1.0, 1.0);
		end
		if (id == MAX_QUEUE_BUTTONS and table.getn(P.OrderedQueue) > MAX_QUEUE_BUTTONS) then
			GameTooltip:AppendText(string.format(FAC_ADDITIONAL_ENTRIES, table.getn(P.OrderedQueue) - MAX_QUEUE_BUTTONS - 1));
			for index = MAX_QUEUE_BUTTONS, table.getn(P.OrderedQueue) do
				local num = P.OrderedQueue[index].numQueued - P.OrderedQueue[index].numProduced;
				GameTooltip:AppendText(num .."x ".. P.OrderedQueue[index].item);
			end
			GameTooltip:AddLine(FAC_CLICK_TO_CANCEL_10);
		else
			local num = P.OrderedQueue[id].numQueued - P.OrderedQueue[id].numProduced;
			GameTooltip:AppendText(num .."x ".. P.OrderedQueue[id].item);
			GameTooltip:AddLine(FAC_CLICK_TO_CANCEL);
		end
		GameTooltip:Show();
	end
end

function P.QueueEverything()
	local numSkills = GetNumTradeSkills();
	local startIndex, stopIndex, increment;
	if (IsAltKeyDown()) then
		startIndex = numSkills;
		stopIndex = 1;
		increment = -1;
	else
		startIndex = 1;
		stopIndex = numSkills;
		increment = 1;
	end
	for index = startIndex, stopIndex, increment do
		local recipeName, _, numAvailable, _, itemName = P.GetTradeSkillInfo(index);
		if (numAvailable > 0) then
			P.AddToQueue(recipeName, itemName, index, numAvailable);
		end
	end
	TradeSkillFrame_SetSelection(GetTradeSkillSelectionIndex());
	TradeSkillFrame_Update();	
	if (table.getn(P.OrderedQueue) == 0) then
		GFWUtils.Print(FAC_NOTHING_TO_QUEUE);
	end
end

-- TODO: option-click for "buy all" equivalent to "queue all"
function P.BuyReagents(amount)
	local index = GetTradeSkillSelectionIndex();
	
	if (IsAltKeyDown()) then
		amount = TradeSkillFrame.numAvailable;
	end
	
	for reagentIndex = 1, GetTradeSkillNumReagents(index) do
		local reagentName, _, reagentCount, playerReagentCount = P.GetTradeSkillReagentInfo(index, reagentIndex);
		local amountNeeded = (reagentCount * amount) - playerReagentCount;
		
		for merchantIndex = 1, GetMerchantNumItems() do
			local name, _, price, numPerBuy, numAvailable, _ = GetMerchantItemInfo(merchantIndex);
			if (name == reagentName) then
				if (numAvailable >= 0 and (numAvailable * numPerBuy) < amountNeeded) then
					amountNeeded = numAvailable * numPerBuy;
				end
				local amountToBuy = math.ceil(amountNeeded / numPerBuy);
				if (amountToBuy > 0) then
					if (numPerBuy > 1) then
						GFWUtils.Print(string.format(FAC_BUYING_SETS_FORMAT, amountToBuy, name));
					else
						GFWUtils.Print(string.format(FAC_BUYING_FORMAT, amountToBuy, name));
					end
					BuyMerchantItem(merchantIndex, math.ceil(amountNeeded / numPerBuy));
				elseif (numAvailable == 0) then
					GFWUtils.Print(string.format(FAC_NOT_AVAILABLE_FORMAT, name));
				else
					GFWUtils.Print(string.format(FAC_ALREADY_ENOUGH_FORMAT, name, amount, P.GetTradeItemName(index)));
				end
				break;
			end
		end
	end
end

function P.Queue(amount)
	local index = GetTradeSkillSelectionIndex();
	local recipeName, _, numAvailable, _, itemName = P.GetTradeSkillInfo(index);

	if (amount > numAvailable) then amount = numAvailable; end

	if (amount > 0) then
		P.AddToQueue(recipeName, itemName, index, amount);
	end
	TradeSkillFrame_SetSelection(GetTradeSkillSelectionIndex());
	TradeSkillFrame_Update();
	P.UpdateDisplay();
end

function P.AddToQueue(recipeName, item, index, amount)

	local alreadyQueued = false;
	GFWUtils.Print(FAC_QUEUEING .. GFWUtils.Hilite(amount .. " x ") .. GetTradeSkillItemLink(index));
	for _, queueEntry in P.OrderedQueue do
		if (item == queueEntry.item) then
			queueEntry.numQueued = queueEntry.numQueued + amount;
			alreadyQueued = true;
		end
	end
	if (not alreadyQueued) then
		local reagentsPerItem = {};
		for reagentIndex = 1, GetTradeSkillNumReagents(index) do
			local reagentName, _, reagentCount, _ = P.GetTradeSkillReagentInfo(index, reagentIndex);
			reagentsPerItem[reagentName] = reagentCount;
		end
		table.insert(P.OrderedQueue, {recipe=recipeName, item=item, numQueued=amount, numProduced=0, icon=GetTradeSkillIcon(index), skill=GetTradeSkillLine(), reagentsPerItem=reagentsPerItem});
	end

	TradeSkillFrame_SetSelection(GetTradeSkillSelectionIndex());
	TradeSkillFrame_Update();
	P.UpdateDisplay();
	if (FAC_Config.AutoIterate) then
		P.ManuallyPaused = false;
		P.Iterate();
	else
		P.ManuallyPaused = true;
	end
	P.UpdateDisplay();
end

function P.UpdateDisplay()
	
	-- we clear these out in case relevant inventory has (actually or effectively) changed.
	P.SkillCache = {};
	P.ReagentCache = {};

	if (MerchantFrame:IsVisible()) then
		AutoCraftBuy:Enable();
	else
		AutoCraftBuy:Disable();
	end
	
	if (table.getn(P.OrderedQueue) > 0) then
		if (P.CraftingQueuedItem) then
			QueueMessageText:Hide();
			QueueMessageText:SetText("");
			AutoCraftRunQueue:SetText(FAC_PAUSE_QUEUE_BUTTON);
		else
			if (not P.ManuallyPaused and P.OrderedQueue[1] ~= nil and P.OrderedQueue[1].skill ~= GetTradeSkillLine()) then
				QueueMessageText:SetText(FAC_QUEUE_PAUSED..": "..string.format(FAC_SWITCH_WINDOW_FORMAT, P.OrderedQueue[1].skill));
			elseif (P.CastingOtherSpell) then
				QueueMessageText:SetText(FAC_QUEUE_PAUSED..": "..string.format(FAC_WAITING_SPELL_FORMAT, CastingBarText:GetText()));
			else
				QueueMessageText:SetText(FAC_QUEUE_PAUSED);
			end
			QueueMessageText:Show();
			AutoCraftRunQueue:SetText(FAC_RUN_QUEUE_BUTTON);
		end

		AutoCraftRunQueue:Enable();
		AutoCraftClear:Enable();
	else
		QueueMessageText:Hide();
		QueueMessageText:SetText("");
		AutoCraftRunQueue:SetText(FAC_RUN_QUEUE_BUTTON);
		AutoCraftRunQueue:Disable();
		AutoCraftClear:Disable();
	end
	
	for i = 1, MAX_QUEUE_BUTTONS do
		button = getglobal("QueuedSkillIcon"..i);
		currentItem = P.OrderedQueue[i];
		if (currentItem == nil) then
			button:Hide();
		else
			if (i == MAX_QUEUE_BUTTONS and table.getn(P.OrderedQueue) > MAX_QUEUE_BUTTONS) then
				SetItemButtonTexture(button, "INV_Misc_QuestionMark");
				SetItemButtonCount(button, (table.getn(P.OrderedQueue) - MAX_QUEUE_BUTTONS - 1));
			else
				SetItemButtonTexture(button, currentItem.icon);
				SetItemButtonCount(button, (currentItem.numQueued - currentItem.numProduced));
			end
			button:Show();
		end
	end
end

function P.StartStop()
	if (P.CraftingQueuedItem == true) then
		P.ManuallyPaused = true;
		SpellStopCasting();
		P.UpdateDisplay();
	else
		P.ManuallyPaused = false;
		P.Iterate();
		P.UpdateDisplay();
	end
end

function P.Clear()
	if (table.getn(P.OrderedQueue) > 0) then
		if (P.CraftingQueuedItem and not P.ManuallyPaused) then
			SpellStopCasting();
		end
		P.OrderedQueue = {};
		TradeSkillFrame_SetSelection(GetTradeSkillSelectionIndex());
		TradeSkillFrame_Update();	
		P.UpdateDisplay();
	end
end

function P.Iterate()	
	if (P.CraftingQueuedItem or P.CastingOtherSpell) then return; end
	if (table.getn(P.OrderedQueue) > 0) then
		local index = P.GetTradeRecipeIndex(P.OrderedQueue[1].recipe);
		if (index >= 0) then
			if (P.OrderedQueue[1].skill == GetTradeSkillLine()) then
				DoTradeSkill(index, (P.OrderedQueue[1].numQueued - P.OrderedQueue[1].numProduced));
				P.UpdateDisplay();
				return;
			end
		end
		P.ManuallyPaused = false;
		P.UpdateDisplay();
	end
end

function P.HandleCreated(item)
	if (P.OrderedQueue[1] and item == P.OrderedQueue[1].item) then
		P.OrderedQueue[1].numProduced = P.OrderedQueue[1].numProduced + 1;
		if (P.OrderedQueue[1].numProduced >= P.OrderedQueue[1].numQueued) then
			P.CraftingQueuedItem = false; -- this will get set back to true when we get the SPELLCAST_START event
			table.remove(P.OrderedQueue, 1);
		end
		TradeSkillFrame_SetSelection(GetTradeSkillSelectionIndex());
		TradeSkillFrame_Update();	
		P.UpdateDisplay();
		--P.Iterate();
		if (table.getn(P.OrderedQueue) > 0) then
			P.AdvanceQueueTime = GetTime() + QUEUE_ADVANCE_DELAY;
		end
	end
end

function P.GetTradeItemName(index)
	local linktext = GetTradeSkillItemLink(index);
	if (linktext) then
		local _,_,item = string.find(linktext, "^.*%[(.*)%].*$");
		return item;
	else
		local item, difficulty, amount = P.Orig_GetTradeSkillInfo(index);
		return item;
	end
end

-- find the trade skill index based on the skill name (recipe name), as indices are subject to change.
function P.GetTradeRecipeIndex(recipeName)
	local numSkills = GetNumTradeSkills();
	for index = 1, numSkills do
		local skillName = P.Orig_GetTradeSkillInfo(index);
		if (skillName == recipeName) then
			return index;
		end
	end
	return -1;
end
